function [u,p,t] = solve_iStokes_Mini(coordinates,elements,dirichlet, ...
                                    neumann,uD,f,g,x0,T,N,theta,nu)
%
% Mini Element (P1 + Bubble)^2 x P1
%
%*** initiate a few constants
nT = size(elements,1);
nC = size(coordinates,1);
%*** First vertex of elements and corresponding edge vectors 
c1 = coordinates(elements(:,1),:);
d21 = coordinates(elements(:,2),:) - coordinates(elements(:,1),:);
d31 = coordinates(elements(:,3),:) - c1;
%*** Vector of element areas |T|
area = abs((d21(:,1).*d31(:,2)-d21(:,2).*d31(:,1))/2);
%*** Assembly of stiffness matrix
telem = [elements,nC+(1:nT)'];
I = reshape(telem(:,[1 2 3 1 2 3 1 2 3 4]),[],1);
J = reshape(telem(:,[1 1 1 2 2 2 3 3 3 4]),[],1);
a = 1/4*(sum(d31.*d31,2)./area);
b = 1/4*(sum(d21.*d31,2)./area);
c = 1/4*(sum(d21.*d21,2)./area);
S = nu * sparse(I,J,[a-2*b+c; b-a; b-c; b-a; a; -b; b-c; -b; c; a-b+c]);
%*** Assembly of mass matrix
I = reshape(telem(:,[1 2 3 4 1 2 3 4 1 2 3 4 1 2 3 4]),[],1);
J = reshape(telem(:,[1 1 1 1 2 2 2 2 3 3 3 3 4 4 4 4]),[],1);
a = 1/12; b = 1/(6*sqrt(10)); c = 1/28;
M = sparse(I,J,area*[2*a a a b a 2*a a b a a 2*a b b b b c]);
%*** Assembly of divergence matrix
I = reshape(telem(:,[1 2 3 4 1 2 3 4 1 2 3 4]),[],1);
J = reshape(telem(:,[1 1 1 1 2 2 2 2 3 3 3 3]),[],1);
a = d31/6; b = d31/(4*sqrt(10)); c = d21/6; d=d21/(4*sqrt(10)); 
B = [c-a;a;-c;b-d;c-a;a;-c;-b;c-a;a;-c;d];    
B = sparse([I;nC+nT+I],[J,J], [B(:,2),-B(:,1)]);  
%*** Computation of free and fixed nodes
dirichlet = unique(dirichlet);
freenodes = setdiff(1:(nC+nT), dirichlet);
freenodes = [freenodes,nC+nT+freenodes,2*(nT+nC)+(1:(nC+isempty(neumann)))];
fixednodes = [dirichlet;nC+nT+dirichlet];
%*** Assembly of right-hand side at t=0
F0 = zeros(2*(nC+nT)+nC,1);
if ~isempty(f)
  fsT = f(coordinates,0);
  F0(1:2*(nC+nT),1) = reshape(M(:,1:nC)*fsT,[],1); 
end
if ~isempty(neumann) 
  cn1 = coordinates(neumann(:,1),:);
  cn2 = coordinates(neumann(:,2),:);
  gmE = g((cn1+cn2)/2,0);
  tmp = repmat((sqrt(sum((cn2-cn1).^2,2))*[1/2,1/2]).*gmE,2,1);
  F0 = F0 + accumarray([neumann(:);nC+nT+neumann(:)],tmp(:),[size(F0,1) 1]);
end  
%*** Loop over timesteps
t = linspace(0,T,N) ;
dt = t(2)-t(1);
u = cell(N,1);
p = cell(N,1);
u{1} = reshape(x0(1:2*(nC+nT)),[],2);
p{1} = x0(2*(nC+nT)+(1:nC));
%*** global matrix
A = [[M+dt*theta*S,sparse(nC+nT,nC+nT);sparse(nC+nT,nC+nT),M+dt*theta*S],...
     -dt*theta*B;-dt*theta*B',sparse(nC,nC)];
D = [[M-dt*(1-theta)*S,sparse(nC+nT,nC+nT);sparse(nC+nT,nC+nT), ...
      M-dt*(1-theta)*S],dt*(1-theta)*B];
if isempty(neumann) 
  C = sparse(elements(:),ones(3*nT,1),area(:,[1,1,1])/3);
  A(2*(nC+nT)+(1:nC),end+1) = C;
  A(end+1,2*(nC+nT)+(1:nC)) = C';
end  
fprintf('       \n');
for k = 1:N-1
  fprintf('\b\b\b\b\b\b\b  %5.3f',k/N);
  %*** Prescribe values at Dirichlet nodes 
  x = zeros(2*(nC+nT)+nC+isempty(neumann),1);
  x(fixednodes) = reshape(uD(coordinates(dirichlet,:),t(k+1)),[],1);
  %*** Assembly of right-hand 
  F = zeros(2*(nC+nT)+nC,1);
  if ~isempty(f)
    fsT = f(coordinates,t(k+1));
    F(1:2*(nC+nT),1) = reshape(M(:,1:nC)*fsT,[],1); 
  end
  if ~isempty(neumann) 
    cn1 = coordinates(neumann(:,1),:);
    cn2 = coordinates(neumann(:,2),:);
    gmE = feval(g,(cn1+cn2)/2);
    tmp = repmat((sqrt(sum((cn2-cn1).^2,2))*[1/2,1/2]).*gmE,2,1);
    F = F + accumarray([neumann(:);nC+nT+neumann(:)],tmp(:),[size(F,1) 1]);
  end 
  %*** assemble non-linear term
  Nu = assembleN0(x0,elements,coordinates); 
  %*** rhs
  b = dt*(theta*F+(1-theta)*F0);
  b(1:2*(nC+nT)) = b(1:2*(nC+nT)) + D*x0(1:3*nC+2*nT);
  b(1:2*(nC+nT)) = b(1:2*(nC+nT)) - dt * (1-theta)* ...
                   Nu(1:2*(nC+nT),1:2*(nC+nT))*x0(1:2*(nC+nT));
  An = A + dt*theta*Nu;
  b = b-An*x;
  x(freenodes) = An(freenodes,freenodes)\b(freenodes);
  u{k+1} = reshape(x(1:2*(nC+nT)),[],2);
  p{k+1} = x(2*(nC+nT)+(1:nC));
  x0 = x;  F0 = F;
end

function Nu = assembleN0(x,elements,coordinates) 
c = 14/sqrt(10);
Ax = zeros(4,4,4);
Ax(2,:,:) = [[ 14,  7,  7, c]; ...
             [  7, 14,  7, c]; ...
             [  7,  7, 14, c]; ...
             [  c,  c,  c, 3]]/168;
Ax(1,:,:) = -Ax(2,:,:);
Ax(4,:,:) = [[ 4*c,    0,  2*c,  3]; ...
             [   0, -4*c, -2*c, -3]; ...
             [ 2*c, -2*c,    0,  0]; ...
             [   3,   -3,    0,  0]]/336;
Ay =  Ax([1,3,2,4],[1,3,2,4],[1,3,2,4]);                          

%*** First vertex of elements and corresponding edge vectors 
d21 = coordinates(elements(:,2),:) - coordinates(elements(:,1),:);
d31 = coordinates(elements(:,3),:) - coordinates(elements(:,1),:);

nT = size(elements,1);
nC = size(coordinates,1);
telem = [elements,nC+(1:nT)'];
A   = zeros(nT,16);
for j=1:4
  %*** a_ij = u * grad(phi_j) * phi_i
  A = A + (x(telem(:,j)).*d31(:,2)-x(nC+nT+telem(:,j)).*d31(:,1))*reshape(Ax(:,:,j),1,[]) ...
        - (x(telem(:,j)).*d21(:,2)-x(nC+nT+telem(:,j)).*d21(:,1))*reshape(Ay(:,:,j),1,[]);
end
II = telem;
JJ = II(:,[1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,4]);
KK = II(:,[1,2,3,4,1,2,3,4,1,2,3,4,1,2,3,4]);    
A = sparse(JJ,KK,A);
Nu = [[A,sparse(nC+nT,nC+nT);sparse(nC+nT,nC+nT),A],sparse(2*(nC+nT),nC); ...
       sparse(nC,3*nC+2*nT)];
     
     
     
     
     
     
     
     
     
     
function [Nu,A] = assembleN(x,elements,coordinates) 
c = 14/sqrt(10);
Ax = zeros(4,4,4);
Ax(2,:,:) = [[ 14,  7,  7, c]; ...
             [  7, 14,  7, c]; ...
             [  7,  7, 14, c]; ...
             [  c,  c,  c, 3]]/168;
Ax(1,:,:) = -Ax(2,:,:);
Ax(4,:,:) = [[ 4*c,    0,  2*c,  3]; ...
             [   0, -4*c, -2*c, -3]; ...
             [ 2*c, -2*c,    0,  0]; ...
             [   3,   -3,    0,  0]]/336;
Ay =  Ax([1,3,2,4],[1,3,2,4],[1,3,2,4]);                          

%*** First vertex of elements and corresponding edge vectors 
d21 = coordinates(elements(:,2),:) - coordinates(elements(:,1),:);
d31 = coordinates(elements(:,3),:) - coordinates(elements(:,1),:);

nT = size(elements,1);
nC = size(coordinates,1);
telem = [elements,nC+(1:nT)'];

tmp   = zeros(nT,16);
% tmp11 = zeros(nT,16);
% tmp21 = zeros(nT,16);
% tmp12 = zeros(nT,16);
% tmp22 = zeros(nT,16);

for j=1:4
  %*** a_ij = u * grad(phi_j) * phi_i
  tmp = tmp + (x(telem(:,j)).*d31(:,2)-x(nC+nT+telem(:,j)).*d31(:,1))*reshape(Ax(:,:,j),1,[]) ...
            - (x(telem(:,j)).*d21(:,2)-x(nC+nT+telem(:,j)).*d21(:,1))*reshape(Ay(:,:,j),1,[]);
%   tmp11 = tmp11 + (x(telem(:,j)).*d31(:,2))*reshape(Ax(j,:,:),1,[]) ...
%                 - (x(telem(:,j)).*d21(:,2))*reshape(Ay(j,:,:),1,[]);
%   tmp12 = tmp12 - (x(telem(:,j)).*d31(:,1))*reshape(Ax(j,:,:),1,[]) ...
%                 + (x(telem(:,j)).*d21(:,1))*reshape(Ay(j,:,:),1,[]);
%   tmp21 = tmp21 + (x(nC+nT+telem(:,j)).*d31(:,2))*reshape(Ax(j,:,:),1,[]) ...
%                 - (x(nC+nT+telem(:,j)).*d21(:,2))*reshape(Ay(j,:,:),1,[]);
%   tmp22 = tmp22 - (x(nC+nT+telem(:,j)).*d31(:,1))*reshape(Ax(j,:,:),1,[]) ...
%                 + (x(nC+nT+telem(:,j)).*d21(:,1))*reshape(Ay(j,:,:),1,[]);
end
II = telem;
JJ = II(:,[1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,4]);
KK = II(:,[1,2,3,4,1,2,3,4,1,2,3,4,1,2,3,4]);    
A = sparse(JJ,KK,tmp);
Nu = [];
% Nu = [A+sparse(JJ,KK,tmp11),sparse(JJ,KK,tmp12); ...
%       sparse(JJ,KK,tmp21),A+sparse(JJ,KK,tmp22)];
%A = [A,sparse(nC+nT,nC+nT);sparse(nC+nT,nC+nT),A];
%size(sparse(nC,3*nC+2*nT))
%size([ [A,sparse(nC+nT,nC+nT);sparse(nC+nT,nC+nT),A],sparse(2*(nC+nT),nC)])
A = [[ [A,sparse(nC+nT,nC+nT);sparse(nC+nT,nC+nT),A],sparse(2*(nC+nT),nC)]; ...
       sparse(nC,3*nC+2*nT)];



